# Generated by Django 2.2.9 on 2020-01-27 18:05

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='LatexConfiguration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('fields', models.TextField(default='')),
                ('model_name', models.CharField(default='', max_length=120)),
                ('template', models.CharField(default='', max_length=200)),
            ],
        ),
    ]
